/*

This source file is part of an archive that may be found at:

    http://www.azillionmonkeys.com/qed/int2.zip

The following code is due to Richard O'Keefe which can be found at:

    http://deja.com/getdoc.xp?fmt=text&AN=338901555

It is claimed to be a natural way for a C programmer to implement a given 2D 
integration algorithm.  An analogous program for Scheme was presented and in a 
head to head time trial it appears as though the Scheme code won by a 
substantial margin; the Scheme code was reportedly 21 times as fast as the C 
code.  There followed a short discussion about this that can be found at:

    http://deja.com/threadmsg_ct.xp?thitnum=22&AN=599815187.1&mhitnum=0&CONTEXT=954640904.534904837

Of course there is a serious problem here.  The Scheme compiler apparently was
targeting a C compiler back end, so we know that there is *some* C code
implementation is competitive to the Scheme code.  So the question remains, is
the C implementation a fair submission?

I claim no.  First of all the code is not ANSI C.  Its not just a little bit
off, its way off.  There is a high degree of unnecessary indirection in the C
code primarily for the purpose of transforming one dimensional function 
parameters into a two dimensional base function called zarkon().  Not 
surprisingly, however, its not so easy to follow the exact sequence of 
zarkon() calls without prior knowledge that this is really just calling the 
function on a rectangular domain.

Lets put some perspective onto this implementation from a language 
capabilities point of view.

On comp.arch Stefan Monnier writes:
 
"From what I can tell, Scheme has reasonably poor support for separate 
compilation, which might be relevant to you or not."

What this means is that passing functions to a set of meta-functions has a
lot less meaning in Scheme than it does in C.  While in the C language,
zarkon() could have been implemented in a completely different file, in a
different context and in fact compiled seperately, (thus justifying the use of 
indirect function pointers), this is not possible in the Scheme world without 
access to the full source code.

This is not just some subtle point, its central to the erroneous way in which
this whole discussion has been presented.  Since it is not possible to saddle
Scheme with the burden of runtime resolved function pointers (and more 
importantly, I don't have a Scheme compiler to try this out on), we instead 
will releave this burden from the C implementations in the examples that 
follow by avoiding function pointers when it is possible.

In none of the examples that follow was any attempt made to simplify the
formula for the function we are calculating.  The function input is a bilinear
function that clearly integrates (numerically as well as continuously) as a 
bi-quadratic.  But the point of the exercise is not to determine how good
of a math student the programmer might be (or in another way we might 
consider this same hoist for different core functions) so this sort of thing
has been ignored.  In call cases, the core function remains in tact, ready to 
be substituted with any other function -- thus retaining the generality of the 
original code.

   -------------------------------------------------------------------------

The following code is not correct ANSI C.  It is reportedly supported by
extensions in the gcc compiler.  I don't have access to a compiler that can
compile it, hence I have not done any direct analysis of it.

double integrate_1d(double l, double u, double (*f)(double)) { 
  double d = (u-l)/8.0;

  return (
   (*f)(l)*0.5+
   (*f)(l+d)+
   (*f)(l+2.0*d)+
   (*f)(l+3.0*d)+
   (*f)(l+4.0*d)+
   (*f)(u-3.0*d)+
   (*f)(u-2.0*d)+
   (*f)(u-d)+
   (*f)(u)*0.5
  )*d;
}

double integrate_2d(double l1, double u1, double l2, double u2,
                    double (*f)(double, double)) { 

  double outer(double y) { 
    double inner(double x) {
      return (*f)(x, y);
    }

    return integrate_1d(l1, u1, inner);
  }

  return integrate_1d(l2, u2, outer);
}

double zark(double u, double v) { 
  double zarkon(double x, double y) {
    return x*y;
  }

  return integrate_2d(0.0, u, 0.0, v, zarkon);
}

*/

/* ------------------------------------------------------------------------- */

// After guessing at what the author might mean (since locally scoped 
// functions is something you just don't see in C programs) I arrived at the
// following reference code.  
//
// The only similarity to the above is the use of the very inner loop.  The 
// rest of the code was arrived at solely as the way I figured a typical C 
// programmer might naturally implement this algorithm.  It is a 2-dimensional 
// 81-sample integration and a doubly nested for() loop is the only way I have 
// even imagined implementing such a thing.

static double zarkonref( double x, double y ) {
    return x*y;
}

static double int2ref( double l1, double u1, double l2, double u2 ) {
double d1, d2, dd, s, t;
double s0, s1, s2, s3;
int i,j;

    d1 = (u1-l1)/8;
    d2 = (u2-l2)/8;
    t = 0;

    for(j=0;j<9;j++,l1+=d1) {

        /* This potential branch mis-predict actually probably leads to a 
           significant performance hit.  Just using a table here would likely 
           lead to a performance increase, however I avoided this level of
           optimization as this might not occurr to a typical C programmer. */

        if( j==0 || j==8 ) dd = 0.5*d2; else dd = d2;

        s = ( zarkonref(l1,l2       )*0.5+
              zarkonref(l1,l2+    d2)+
              zarkonref(l1,l2+2.0*d2)+
              zarkonref(l1,l2+3.0*d2)+
              zarkonref(l1,l2+4.0*d2)+
              zarkonref(l1,u2-3.0*d2)+
              zarkonref(l1,u2-2.0*d2)+
              zarkonref(l1,u2-    d2)+
              zarkonref(l1,u2       )*0.5);

        t += s * dd;
    }

    return t*d1;
}

static double zarkref( double u, double v ) {
   return int2ref(0,u,0,v);
}

/* ------------------------------------------------------------------------- */

// Ok, so lets say we object to the above implementation since the otherwise
// symmetrical summation appears to be implemented in two different ways.
// In the following I have basically hacked something which is as close to
// what I think is meant by the original code.  It is conservative by virtue 
// of using static globals as a means of passing beyond one level of scope -- 
// the C compiler has fewer opportunities to perform parameter optimizations
// than would be available in a compiler/language that supports the original
// idea.  However, it is also at least partially aggressive in that, like the
// other routines, I have at least removed unnecessary indirection to the
// inner most loop.
//
// As I originally suspected there are far too many levels of indirection for
// the compiler to ideally inline the code.  However, this certainly does not
// justify the original claim, as this is by no means a natural implementation
// by function pointers.  In particular, knowing that there are no erroneous
// side effects in these global statics is not so easy to see.  This is 
// analogous to trying to figure out exactly what parameters are being passed
// from top to bottom of the original function.
//
// The performance should be very similar to the speed of the originally
// intended code, with the exception that "inner()" does not call a function
// pointer, but rather calls the function directly.

static double integrate_1d(double l, double u, double (*f)(double)) { 
  double d = (u-l)/8.0;

  return (
    f(l)*0.5+
    f(l+d)+
    f(l+2.0*d)+
    f(l+3.0*d)+
    f(l+4.0*d)+
    f(u-3.0*d)+
    f(u-2.0*d)+
    f(u-d)+
    f(u)*0.5
  )*d;
}

static double zarkon( double x, double y ) {
  return x*y;
}

static double inner_y;

static double inner(double x) {
  return zarkon(x, inner_y);
}

static double outer_l1;
static double outer_u1;

static double outer(double y) { 
  inner_y = y;
  return integrate_1d( outer_l1, outer_u1, inner );
}

static double integrate_2d(double l1, double u1, double l2, double u2) {

  outer_l1 = l1;
  outer_u1 = u1;

  return integrate_1d(l2, u2, outer);
}

double zark( double u, double v ) {
  return integrate_2d(0.0, u, 0.0, v);
}

/* ------------------------------------------------------------------------- */

// The following implementation tries to preserve the original spirit of the
// scheme implementation by having only written mechanism for performing the
// linear integration sum.  However, input parameters are taken care of by a
// combination of macros and simple function calls that the compiler can
// inline with suitable optimization setting.
//
// Following the parameters is a little complicated, but represents, in final 
// post-processed code, the same control flow of the parameters in the same
// order as the original code implied.

#define mac_int1d(xl,xu,xd,yl,yu,f) \
   (  f(yl,yu,xl,    0)*0.5 +       \
      f(yl,yu,xl,   xd) +           \
      f(yl,yu,xl, 2*xd) +           \
      f(yl,yu,xl, 3*xd) +           \
      f(yl,yu,xl, 4*xd) +           \
      f(yl,yu,xu,-3*xd) +           \
      f(yl,yu,xu,-2*xd) +           \
      f(yl,yu,xu,  -xd) +           \
      f(yl,yu,xu,    0)*0.5)*xd

#define mac_inner(y,dummy,x,xd) zarkon(x+xd,y)

// Rather than macro-izing the mac_outer() function, it is simply a trivial
// function call.  The reason is that the macro expansion rules in WATCOM 
// C/C++ stops at recusion.  I am not sure whether or not this is a required 
// behaviour in ANSI C, however it fairly clear why WATCOM implemented it this 
// way.

static double mac_outer(double xl,double xu,double y,double yd) {
    double xd = (xu-xl)/8;

    return mac_int1d(xl,xu,xd,y+yd,0,mac_inner);
}

static double mac_int2d(double l1, double u1, double l2, double u2) {
    double yd = (u2-l2)/8;

    return mac_int1d(l2,u2,yd,l1,u1,mac_outer);
}

static double mac_zark( double u, double v ) {
    return mac_int2d(0.0, u, 0.0, v);
}

/* ------------------------------------------------------------------------- */

// x86 processor specific timing functions.  Requires a Pentium of later CPU.

static double tsc(void) {
double retval;

    _asm {
        .586p
        rdtsc
        mov dword ptr retval[0],eax
        mov dword ptr retval[4],edx
        fild qword ptr retval
        fstp retval
    }

    return retval;
}

double TimeTrial( double (*f) (double, double) ) {
double t,s;
int c=0;

    // Under Windows NT the "cli" needs to be omitted.
    _asm { 
        .586p
        cli
        cpuid
    }

    t = tsc();

    for(c=0,s=3;c<10;c++,s+=0.1) {
        f( s, s+1.75 );
        f( s, s+1.75 );
        f( s, s+1.75 );
        f( s, s+1.75 );
        f( s, s+1.75 );

        f( s, s+1.75 );
        f( s, s+1.75 );
        f( s, s+1.75 );
        f( s, s+1.75 );
        f( s, s+1.75 );
    }

    t = tsc()-t;

    // Under Windows NT the "sti" needs to be omitted.
    _asm { sti }

    return t;
}

/* ------------------------------------------------------------------------- */

#include <stdio.h>

main() {

    /* These are here just to test the correctness */

    printf( "%g\n", zarkref( 5, 6.3 ) );
    printf( "%g\n", zark( 5, 6.3 ) );
    printf( "%g\n", mac_zark( 5, 6.3 ) );

    /* Here are the actual time trials.  By walking through the zark() code, 
       it appears as though there are 167 adds and 101 muls.  On the Athlon 
       CPU, on which this test was performed, the peak performance is limited 
       by ther maximum of either the adds or muls.  So in this case the 167 
       adds.  The TimeTrial() routine performs 100 such integrations and on
       top of that performs 110 additions, so the total number of add-ops is
       16810.  Since the Athlon cannot perform more than one FP add per clock,
       the number of clocks per FP add represents the efficiency versus the
       "speed of light" code that might be achieved by hand coded assembly.
    */

    printf( "Time: %g cycles/add-op\n", TimeTrial(zarkref)/16810 );
    printf( "Time: %g cycles/add-op\n", TimeTrial(zark)/16810 );
    printf( "Time: %g cycles/add-op\n", TimeTrial(mac_zark)/16810 );
}

/* ------------------------------------------------------------------------- */

// Results
//
// The following results we achieved using the WATCOM C/C++ 11.0 compiler on 
// an AMD Athlon running at 500Mhz.  The optimizations options chose were: 
// "/6r /s /otxan /oe=999" which roughly means: "Pentium Pro style register 
// usage", "No stack checking", "optimize for speed", "unroll loops", "replace 
// divides by reciprocal multiply", and "aggressively inline".
// 
// Nested loop:                2.9326 cycles/add-op
// Pascal/Scheme-like code:   11.2176 cycles/add-op
// Macro hybrid:               2.8313 cycles/add-op
// Hand teaked asm:            1.0000 cycles/add-op  (theoretical)
//
// From eyeballing a few runs, the timer appears to have a maxmim error of 
// about 3%. So at the very least the original claim that Scheme was 21 times 
// faster than the natural solution in C is not very credible. Even in the 
// worst situation the best result that the Scheme solution might achieve is 
// 11 times the performance of a poor C implementation, and that's only if the
// Scheme solution gave the absolutely idea level of performance.
//
// For the more "natural" C implementations that margin is reduced to about
// 3 times.  Is it even possible that Scheme will come closer to the speed
// of light performance than C given a margin of only 3x?
//
// I examined the disassembly of these routines and found that the inner loops
// of the two faster routines contained almost no control or parameter passing
// overhead.  In other words, the time lost versus the theoretical speed of
// light implementation is most probably due entirely to scheduling and/or x87 
// stack manipulation overhead.
//
// Looking back at the original code sequences, though, this was a clearly a
// question about flow control, not FPU scheduling capabilities.  Given that
// the original Scheme compiler targets a C compiler -- it is highly unlikely
// that it could produce significantly better scheduled code than the original
// C solution.  Unfortunately, I do not have a Scheme compiler with which to
// complete this exercise.  
//
// However, I believe that there is sufficient evidence to suggest that at 
// least in this example, Scheme does not have any measureable performance 
// advantage over how a typical C programmer might naturally program something
// like this.
